"""Pieces for templating multiple dashboards at once."""
import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional, Union

from grafana_client import GrafanaApi

from grafanarmadillo.alerter import Alerter
from grafanarmadillo.dashboarder import Dashboarder
from grafanarmadillo.find import Finder
from grafanarmadillo.templator import Templator
from grafanarmadillo.util import resolve_object_to_filepath


class Store(ABC):
	"""A destination or source for items."""

	@abstractmethod
	def read_alert(self, name):
		"""Read an alert from this store."""

	@abstractmethod
	def read_dashboard(self, name):
		"""Read a dashboard from this store."""

	@abstractmethod
	def write_alert(self, name, alert):
		"""Write an alert to this store."""

	@abstractmethod
	def write_dashbaord(self, name, dashboard):
		"""Write an alert to this store."""


@dataclass
class FileStore(Store):
	"""
	Store and retrieve Grafana objects in the filesystem.

	Objects will be stored under the same path as in Grafana.
	For example, a dashboard titled "MyDashboard" in a folder titled "MyFolder"
	will appear at `{root}/MyFolder/MyDashboard.json`.
	"""

	root: Path

	@staticmethod
	def _read(file: Path) -> dict:
		with file.with_suffix(".json").open(mode="r", encoding="utf-8") as f:
			return json.load(f)

	@staticmethod
	def _write(file: Path, content: dict):
		file.parent.mkdir(exist_ok=True)
		with file.with_suffix(".json").open(mode="w", encoding="utf-8") as f:
			json.dump(content, f)

	def read_alert(self, name):
		"""Read an alert from this store."""
		return self._read(resolve_object_to_filepath(self.root, name))

	def read_dashboard(self, name):
		"""Read a dashboard from this store."""
		return self._read(resolve_object_to_filepath(self.root, name))

	def write_alert(self, name, alert):
		"""Write an alert to this store."""
		return self._write(resolve_object_to_filepath(self.root, name), alert)

	def write_dashbaord(self, name, dashboard):
		"""Write an alert to this store."""
		return self._write(resolve_object_to_filepath(self.root, name), dashboard)


@dataclass
class GrafanaStore(Store):
	"""Store and retrieve objects from a Grafana instance."""

	gfn: GrafanaApi

	def read_alert(self, name):
		"""Read an alert from this store."""
		finder, alerter = Finder(self.gfn), Alerter(self.gfn)
		alert_info, _ = finder.create_or_get_alert(name)
		alert, _ = alerter.export_alert(alert_info)
		return alert_info

	def read_dashboard(self, name):
		"""Read a dashboard from this store."""
		finder, dashboarder = Finder(self.gfn), Dashboarder(self.gfn)
		dashboard_info, _ = finder.create_or_get_dashboard(name)
		dashboard_content, _ = dashboarder.export_dashboard(dashboard_info)
		return dashboard_content

	def write_alert(self, name, alert):
		"""Write an alert to this store."""
		finder, alerter = Finder(self.gfn), Alerter(self.gfn)
		alert_info, folder_info = finder.create_or_get_alert(name)
		alerter.import_alert(alert, folder_info)

	def write_dashbaord(self, name, dashboard):
		"""Write an alert to this store."""
		finder, dashboarder = Finder(self.gfn), Dashboarder(self.gfn)
		dashboard_info, folder = finder.create_or_get_dashboard(name)
		dashboarder.import_dashboard(dashboard, folder)


@dataclass
class Alert:
	"""Flowable request for an Alert."""

	name_obj: str
	name_tmpl: str
	templator: Templator


@dataclass
class Dashboard:
	"""Flowable request for a Dashboard."""

	name_obj: str
	name_tmpl: str
	templator: Templator


Flowable = Union[Alert, Dashboard]


class FlowException(Exception):
	"""Wrapped Exception of running a Flow."""

	def __init__(self, item: Flowable, cause: Optional[BaseException] = None):
		self.item = item
		if cause:
			self.__cause__ = cause


@dataclass
class FlowResult:
	"""Result of running a Flow."""

	successes: List[Flowable]
	failures: List[FlowException]

	def raise_first(self):
		"""Raise the first exception, if present."""
		if self.failures:
			raise self.failures[0]


@dataclass
class Flow:
	"""A collection of templating actions to do."""

	store_obj: Store
	store_tmpl: Store
	flows: List[Flowable] = field(default=list)

	def append(self, flow: Flowable):
		"""Add a Flowable request to this Flow."""
		self.flows.append(flow)

	def obj_to_tmpl(self) -> FlowResult:
		"""Import from the source to the destination."""
		return self.run(obj_to_tmpl=True)

	def tmpl_to_obj(self) -> FlowResult:
		"""Export from the destination to the source."""
		return self.run(obj_to_tmpl=False)

	def run(self, obj_to_tmpl: bool) -> FlowResult:
		"""Run the flow."""
		failures = []
		successes = []

		for item in self.flows:
			try:
				if isinstance(item, Alert):
					if obj_to_tmpl:
						obj = self.store_obj.read_alert(item.name_obj)
						tmpl = item.templator.make_template_from_dashboard(obj)
						self.store_tmpl.write_alert(item.name_tmpl, tmpl)
					else:
						tmpl = self.store_tmpl.read_alert(item.name_tmpl)
						info = self.store_obj.read_alert(item.name_obj)
						obj = item.templator.make_dashboard_from_template(info, tmpl)
						self.store_obj.write_alert(item.name_obj, obj)
				elif isinstance(item, Dashboard):
					if obj_to_tmpl:
						obj = self.store_obj.read_dashboard(item.name_obj)
						tmpl = item.templator.make_template_from_dashboard(obj)
						self.store_tmpl.write_dashbaord(item.name_tmpl, tmpl)
					else:
						tmpl = self.store_tmpl.read_dashboard(item.name_tmpl)
						info = self.store_obj.read_dashboard(item.name_obj)
						obj = item.templator.make_dashboard_from_template(info, tmpl)
						self.store_obj.write_dashbaord(item.name_obj, obj)
				else:
					raise TypeError(
						f"Invalid flow, expected one of {Alert.__name__}, {Dashboard.__name__}, received {item.__class__.__name__}")

				successes.append(item)
			except Exception as e:
				failures.append(FlowException(item, e))

		return FlowResult(successes, failures)



"""
Perform bulk operations on Grafana.

BulkOperations defines the basic interface. It is meant to be parametrised in 2 phases:
- information source: Subclasses of BulkOperations implement methods to get orgs and resources
- actions: Subclasses of those classes implement the actions to take

For example:
	BulkGrafanaOperation uses a Grafana instance as its source
	BulkExporter uses BulkGrafanaOperations to list all objects and write them to disk
"""
import logging
from abc import ABC, abstractmethod
from functools import lru_cache
from pathlib import Path
from typing import Generator, List, Tuple

from grafana_client import GrafanaApi

from grafanarmadillo.alerter import Alerter
from grafanarmadillo.dashboarder import Dashboarder
from grafanarmadillo.find import Finder
from grafanarmadillo.paths import PathCodec
from grafanarmadillo.templator import Templator
from grafanarmadillo.types import AlertContent, DashboardContent, GrafanaPath, OrgMeta
from grafanarmadillo.util import exactly_one, read_from_file, write_to_file


l = logging.getLogger(__name__)


@lru_cache
def get_all_orgs(gfn_multiorg) -> List[OrgMeta]:
	"""Get a list of all orgs."""
	return gfn_multiorg.organizations.list_organization()


@lru_cache
def get_org(gfn_multiorg, org_name: str):
	"""Get an org by name, efficiently."""
	return exactly_one(
		list(filter(lambda o: o["name"] == org_name, get_all_orgs(gfn_multiorg))),
		f"org with name {org_name}"
	)


class BulkOperation(ABC):
	"""Run bulk operations on Grafana."""

	def __init__(self, cfg: dict):
		self.cfg = cfg
		self.gfn_multiorg = GrafanaApi(**self.cfg)

	def run(self):
		"""Run this bulk operation."""
		for org, gfn in self.all_orgs():
			for out_path, dashboard_content in self.get_all_dashboards(org, gfn):
				self.each_dashboard(out_path, dashboard_content)
			for out_path, alert_content in self.get_all_alerts(org, gfn):
				self.each_alert(out_path, alert_content)

	@abstractmethod
	def all_orgs(self) -> Generator[Tuple[OrgMeta, GrafanaApi], None, None]:
		"""
		Iterate over all organisations.

		This method should be implemented for each source of information.
		"""

	@abstractmethod
	def get_all_dashboards(self, org: OrgMeta, gfn: GrafanaApi) -> Generator[Tuple[GrafanaPath, DashboardContent], None, None]:
		"""
		Iterate over all dashboards.

		This method should be implemented for each source of information.
		"""

	@abstractmethod
	def get_all_alerts(self, org: OrgMeta, gfn: GrafanaApi) -> Generator[Tuple[GrafanaPath, AlertContent], None, None]:
		"""
		Iterate over all alerts.

		This method should be implemented for each source of information.
		"""

	@abstractmethod
	def each_dashboard(self, path: GrafanaPath, dashboard: DashboardContent):
		"""
		Act on each dashboard in Grafana.

		This method should be implemented for each operation.
		"""

	@abstractmethod
	def each_alert(self, path: GrafanaPath, alert: AlertContent):
		"""
		Act on each alert in Grafana.

		This method should be implemented for each operation.
		"""


class BulkGrafanaOperation(BulkOperation, ABC):
	"""Bulk operations which uses a Grafana instance as its source."""

	def all_orgs(self) -> Generator[Tuple[OrgMeta, GrafanaApi], None, None]:
		"""Iterate over all organisations in Grafana."""
		orgs = self.gfn_multiorg.organizations.list_organization()
		for org in orgs:
			gfn = GrafanaApi(**{**self.cfg, "organization_id": org["id"]})
			yield org, gfn

	def get_all_dashboards(self, org: OrgMeta, gfn: GrafanaApi) -> Generator[Tuple[GrafanaPath, DashboardContent], None, None]:
		"""Get all dashboards."""
		finder, dashboarder = Finder(gfn), Dashboarder(gfn)
		dashboards = finder.list_dashboards()
		for dashboard in dashboards:
			dashboard_content, folder = dashboarder.export_dashboard(dashboard)

			if folder is None:
				folder_name = "General"
			else:
				folder_name = folder["name"]

			out_path = GrafanaPath(dashboard_content["title"], folder_name, org["name"])

			yield out_path, dashboard_content

	def get_all_alerts(self, org: OrgMeta, gfn: GrafanaApi) -> Generator[Tuple[GrafanaPath, AlertContent], None, None]:
		"""Get all alerts."""
		finder, alerter = Finder(gfn), Alerter(gfn)
		alerts = finder.list_alerts()
		for alert in alerts:
			alert_content, folder = alerter.export_alert(alert)

			folder_name = folder["title"]
			out_path = GrafanaPath(alert_content["title"], folder_name, org["name"])

			yield out_path, alert_content


class BulkFileOperation(BulkOperation, ABC):
	"""Bulk operation which uses a filetree as its source."""

	def __init__(self, cfg: dict, root_directory: Path):
		self.root_directory = root_directory
		super().__init__(cfg)

	def all_orgs(self) -> Generator[Tuple[OrgMeta, GrafanaApi], None, None]:
		"""Iterate over all organisations in Grafana."""
		orgs = {PathCodec.decode_segment(o.name) for o in self.root_directory.glob("*/*")}
		for org_name in orgs:
			org = self.gfn_multiorg.organization.find_organization(org_name)
			gfn = GrafanaApi(**{**self.cfg, "organization_id": org["id"]})

			yield org, gfn

	def get_all_dashboards(self, org: OrgMeta, gfn: GrafanaApi) -> Generator[Tuple[GrafanaPath, DashboardContent], None, None]:
		"""Get all dashboards."""
		folders = (self.root_directory / "dashboards" / PathCodec.encode_segment(org["name"])).glob("*")
		for folder_path in folders:
			for dashboard_path in folder_path.glob("*.json"):
				content = read_from_file(dashboard_path)
				yield GrafanaPath(PathCodec.decode_segment(dashboard_path.name), PathCodec.decode_segment(folder_path.name), org["name"]), content

	def get_all_alerts(self, org: OrgMeta, gfn: GrafanaApi) -> Generator[Tuple[GrafanaPath, AlertContent], None, None]:
		"""Get all alerts."""
		folders = (self.root_directory / "alerts" / PathCodec.encode_segment(org["name"])).glob("*")
		for folder_path in folders:
			for alert_path in folder_path.glob("*.json"):
				content = read_from_file(alert_path)
				yield GrafanaPath(PathCodec.decode_segment(alert_path.name), PathCodec.decode_segment(folder_path.name), org["name"]), content


class BulkExporter(BulkGrafanaOperation):
	"""Export all resources from Grafana to files."""

	def __init__(self, cfg: dict, root_directory: Path, templator: Templator):
		self.root_directory = root_directory
		self.templator = templator
		super().__init__(cfg)

	def each_dashboard(self, path: GrafanaPath, dashboard: DashboardContent):
		"""Write each dashboard to files."""
		dashboard_templated = self.templator.make_template_from_dashboard(dashboard)
		l.info(f"export dashboard path={path}")
		write_to_file((self.root_directory / "dashboards" / PathCodec.encode_grafana(path)).with_suffix(".json"), dashboard_templated)

	def each_alert(self, path: GrafanaPath, alert: AlertContent):
		"""Write each alert to files."""
		alert_templated = self.templator.make_template_from_dashboard(alert)
		l.info(f"export alert path={path}")
		write_to_file((self.root_directory / "alerts" / PathCodec.encode_grafana(path)).with_suffix(".json"), alert_templated)


class BulkImporter(BulkFileOperation):
	"""Import all resources from files into Grafana."""

	def __init__(self, cfg: dict, root_directory: Path, templator: Templator):
		self.templator = templator
		super().__init__(cfg, root_directory)

	def each_dashboard(self, path: GrafanaPath, dashboard: DashboardContent):
		"""Import each dashboard into Grafana."""
		org = get_org(self.gfn_multiorg, path.org)
		gfn = GrafanaApi(**{**self.cfg, "organization_id": org["id"]})

		finder, dashboarder = Finder(gfn), Dashboarder(gfn)
		folder = finder.create_or_get_folder(path.folder)
		dashboard_templated = self.templator.make_dashboard_from_template(
			dashboard, dashboard
		)
		l.info(f"import dashboard path={path}")
		dashboarder.import_dashboard(dashboard_templated, folder)

	def each_alert(self, path: GrafanaPath, alert: AlertContent):
		"""Import each alert into Grafana."""
		org = get_org(self.gfn_multiorg, path.org)
		gfn = GrafanaApi(**{**self.cfg, "organization_id": org["id"]})

		finder, alerter = Finder(gfn), Alerter(gfn)
		folder = finder.create_or_get_folder(path.folder)
		alert_templated = self.templator.make_dashboard_from_template(alert, alert)
		l.info(f"import alert path={path}")
		alerter.import_alert(alert_templated, folder)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""Setup dot py."""
from __future__ import absolute_import, print_function

import re
# import re
from glob import glob
from os.path import basename, dirname, join, splitext

from setuptools import find_packages, setup


def read(*names, **kwargs):
	"""Read description files."""
	path = join(dirname(__file__), *names)
	with open(path, encoding=kwargs.get("encoding", "utf8")) as fh:
		return fh.read()


long_description = "{}\n{}".format(
	read("README.rst"), read(join("docs", "CHANGELOG.rst")),
)
long_description = re.sub(r":doc:`(.*)`", r"https://github.com/lilatomic/grafanarmadillo/tree/main/docs/rst/\1.rst", long_description)

setup(
	name="grafanarmadillo",
	version="0.6.0",
	description="Simplifies interacting with Grafana, with a focus on templating dashboards",
	long_description=long_description,
	long_description_content_type="text/x-rst",
	license="MIT License",
	author="lilatomic",
	url="https://github.com/lilatomic/grafanarmadillo",
	packages=find_packages("src"),
	package_dir={"": "src"},
	py_modules=[splitext(basename(i))[0] for i in glob("src/*.py")],
	include_package_data=True,
	zip_safe=False,
	classifiers=[],
	project_urls={
		"webpage": "https://github.com/lilatomic/grafanarmadillo",
		"Documentation": "https://grafanarmadillo.readthedocs.io/en/latest/",
		"Changelog": "https://github.com/lilatomic/grafanarmadillo/blob/main/docs/CHANGELOG.rst",
		"Issue Tracker": "https://github.com/lilatomic/grafanarmadillo/issues",
		"Discussion Forum": "https://github.com/lilatomic/grafanarmadillo/discussions",
	},
	keywords=[],
	python_requires=">=3.8, <4",
	install_requires=["grafana-client~=3.0"],
	extras_require={
		'cli': ["click>=8"],
		'migrate': ["docker>=5"]
	},
	setup_requires=[
		#   'pytest-runner',
		#   'setuptools_scm>=3.3.1',
	],
	entry_points={
		"console_scripts": ["grafanarmadillo=grafanarmadillo.cmd:grafanarmadillo"]
		#
	},
)



"""Tools for manipulating path-like objects into references to Grafana objects or file-safe paths."""
from pathlib import Path
from typing import List
from urllib.parse import quote_plus, unquote_plus

from grafanarmadillo.types import GrafanaPath


class PathCodec:
	"""Safely encode paths which may contain invalid characters, such as forward slashes."""

	@staticmethod
	def encode_grafana(path: GrafanaPath) -> Path:
		"""Encode a GrafanaPath."""
		return PathCodec.encode([path.org, path.folder, path.name])

	@staticmethod
	def encode(segments: List[str]) -> Path:
		"""Encode segments to a path."""
		encoded_segments = [quote_plus(segment) for segment in segments]
		return Path(*encoded_segments)

	@staticmethod
	def encode_segment(segment: str) -> str:
		"""Encode a single segment."""
		return quote_plus(segment)

	@staticmethod
	def decode_grafana(path: Path) -> GrafanaPath:
		"""Decode a Path to a GrafanaPath."""
		parts = PathCodec.decode(path)

		if len(parts) == 3:
			return GrafanaPath(org=parts[0], folder=parts[1], name=parts[2])
		elif len(parts) == 2:
			return GrafanaPath(folder=parts[0], name=parts[1])
		elif len(parts) == 1:
			return GrafanaPath(folder="General", name=parts[0])
		else:
			raise ValueError(f"Grafana path has too many parts {path=} len={len(parts)}")

	@staticmethod
	def decode(path: Path) -> List[str]:
		"""Decode a path to segments."""
		parts = path.parts[1:] if path.is_absolute() else path.parts

		decoded_segments = [unquote_plus(segment) for segment in parts]
		return decoded_segments

	@staticmethod
	def decode_segment(segment: str) -> str:
		"""Decode a single segment."""
		return unquote_plus(segment)
